@@ -10,7 +10,7 @@ import traceback
 
 from azure.storage.queue import QueueClient
 from azure.storage.blob import BlobClient
-from azure.cosmosdb.table import TableService
+from azure.cosmosdb.table import TableService, TableBatch
 
 enough_workers_message = "There are already enough active workers"
 workers_long_running = "Workers are active for too long. Going to process as well."
@@ -143,10 +143,18 @@ class Worker:
             else:
                 raise
 
+    # There are 3 scenarious:
+    # 1. Small blob (under segment_limit size) is being processed. Result is written to the output blob. Task state should be "Completed"
+    # 2. Large blob (more than segment_limit) is being processed. Partial result is written to blob_0. Result of blob splitted into segment_list size parts and enqueued.
+    # 3. Parent task (with child_task_partition field). We poll child tasks. Record progress. If child task fails -> parent fails as well.
     def _process_blob(self, request_id, request_entity, func):
-        blob_name, params, offset, retry_count, batch_count = self._read_metadata(request_entity)
+        blob_name, params, child_task_partition, retry_count = self._read_metadata(request_entity)
+        if child_task_partition:
+            self._poll_child_tasks(request_entity, request_id, child_task_partition)
+            return
+
         try:
-            result, whole_blob_read = self._process_blob_by_segments(blob_name, func, offset, params)
+            result, whole_blob_read, blobs_iterator = self._process_blob_by_segments(blob_name, func, params)
         except Exception as e:
             if retry_count == 3:
                 self._record_failure(request_id, e)
@@ -155,28 +163,93 @@ class Worker:
                 self.table_client.update_entity(self.metadata_table_name, request_entity)
                 self._enqueue(request_id)
             return
-            
+        
         serialized_result = "\n".join(json.dumps(r) for r in result)
-        if not whole_blob_read:
-            out_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch_count))
+        if whole_blob_read:
+            out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
             out_blob.upload_blob(serialized_result, timeout=1200)
-                
-            request_entity['Offset'] = offset + len(result)
-            request_entity['Batch'] = batch_count + 1
+            request_entity['State'] = 'Completed'
+            self.table_client.update_entity(self.metadata_table_name, request_entity)
+        else:
+            out_blob = self._get_output_blob("%s/%s_0" % (self.worker_name, blob_name))
+            out_blob.upload_blob(serialized_result, timeout=1200)
+            blob_names = self._split_blob_into_chunks(blobs_iterator, blob_name)
+            partition_key = self._enqueue_child_tasks(blob_names, params)
+            request_entity['ChildTasksPartition'] = partition_key
+            request_entity['State'] = 'Polling'
             self.table_client.update_entity(self.metadata_table_name, request_entity)
             self._enqueue(request_id)
+
+    def _poll_child_tasks(self, request_entity, request_id, child_tasks_partition):
+        completed_blobs = []
+        total_count = 0
+        for task_entity in self.table_client.query_entities(self.metadata_table_name, filter="PartitionKey eq '%s'" % child_tasks_partition):
+            total_count += 1
+            if 'State' in task_entity and task_entity['State'] == 'Completed':
+                completed_blobs.append(task_entity["BlobName"])
+        if len(completed_blobs) == total_count:
+            self._merge_blobs(completed_blobs, request_entity["BlobName"])
+            request_entity["State"] = "Completed"
+            self.table_client.update_entity(self.metadata_table_name, request_entity)
         else:
-            out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
-            if offset == 0:
+            request_entity['Progress'] = len(completed_blobs) * self.segment_limit
+            self.table_client.update_entity(self.metadata_table_name, request_entity)
+            self._enqueue(request_id)
+
+    def _split_blob_into_chunks(self, segment_iterator, blob_name):
+        blob_names = []
+        batch = []
+        batch_number = 1
+        for source_segment in segment_iterator:
+            batch.append(source_segment)
+            if len(batch) == self.segment_limit:
+                partial_blob_name = "%s/%s_%d" % (self.worker_name, blob_name, batch_number)
+                out_blob = self._get_input_blob(partial_blob_name)
+                serialized_result = json.dumps(batch)
                 out_blob.upload_blob(serialized_result, timeout=1200)
-            else:     
-                content = []
-                for batch in range(batch_count):
-                    in_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch))
-                    blob_content = in_blob.download_blob().readall().decode()
-                    content.append(blob_content)
-                content.append(serialized_result)
-                out_blob.upload_blob("\n".join(content), timeout=1200)
+                blob_names.append(partial_blob_name)
+                batch = []
+                batch_number += 1
+        if len(batch):
+            partial_blob_name = "%s/%s_%d" % (self.worker_name, blob_name, batch_number)
+            out_blob = self._get_input_blob(partial_blob_name)
+            serialized_result = json.dumps(batch)
+            out_blob.upload_blob(serialized_result, timeout=1200)
+            blob_names.append(partial_blob_name)
+        return blob_names
+
+    def _enqueue_child_tasks(self, blob_names, params):
+        partition_key = str(uuid.uuid4()).replace("-", "")
+        request_hashes = [str(hashlib.md5(data.encode()).hexdigest()) for data in blob_names]
+        request_ids = ["%s_%s" % (partition_key, request_hash) for request_hash in request_hashes]
+        batch = TableBatch()
+        batch_size = 0
+        for partial_blob_name, request_hash in zip(blob_names, request_hashes):
+            metadata_entity = {"PartitionKey": partition_key, "RowKey": request_hash, "BlobName": partial_blob_name }
+            if params:
+                metadata_entity['Parameters'] = json.dumps(params)
+            batch.insert_entity(metadata_entity)
+            batch_size += 1
+            if batch_size == 100:
+                self.table_client.commit_batch(self.metadata_table_name, batch)
+                batch_size = 0
+                batch = TableBatch()
+                
+        if batch_size > 0:
+            self.table_client.commit_batch(self.metadata_table_name, batch)
+
+        for child_request_id in request_ids:
+            self._enqueue(child_request_id)
+        return partition_key
+
+    def _merge_blobs(self, blob_names, blob_name):
+        out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_names))
+        content = []
+        for partial_blob_name in sorted(blob_names, key=lambda b: int(b.split('_')[-1])):
+            in_blob = self._get_output_blob("%s/%s" % (self.worker_name, partial_blob_name))
+            blob_content = in_blob.download_blob().readall().decode()
+            content.append(blob_content)
+        out_blob.upload_blob("\n".join(content), timeout=1200)
 
     def _get_input_blob(self, blob_name):
         return BlobClient.from_connection_string(self.ltb_connection_string, container_name=input_container, blob_name=blob_name)
@@ -190,40 +263,34 @@ class Worker:
         if 'Parameters' in request_entity:
             params = json.loads(request_entity['Parameters'])
         offset = 0
-        if 'Offset' in request_entity:
-            offset = request_entity['Offset']
+        if 'ChildTasksPartition' in request_entity:
+            child_tasks_partition = request_entity['ChildTasksPartition']
         retry_count = 0
         if 'RetryCount' in request_entity:
             retry_count = request_entity['RetryCount']
-        batch_count = 0
-        if 'Batch' in request_entity:
-            batch_count = request_entity['Batch']
-        return blob_name, params, offset, retry_count, batch_count
+        return blob_name, params, child_tasks_partition, retry_count
 
-    def _process_blob_by_segments(self, blob_name, func, offset, params):
+    def _process_blob_by_segments(self, blob_name, func, params):
         result = []
         in_blob = self._get_input_blob(blob_name)
         batch = []
-        segment_id = -1
         whole_blob_read = True
-        for source_segment in self._read_blob_by_segments(in_blob):
-            segment_id += 1
-            if segment_id < offset:
-                continue
-
+        segments_processed = 0
+        blobs_iterator = self._read_blob_by_segments(in_blob)
+        for source_segment in blobs_iterator:
             batch.append(source_segment)
             if len(batch) == self.batch_size:
                 result.extend(func(batch, params))
+                segments_processed += len(batch)
                 batch = []
+                if segments_processed >= self.segment_limit:
+                    whole_blob_read = False
+                    break
 
-            if segment_id - offset + 1 >= self.segment_limit:
-                whole_blob_read = False
-                break
-
-        if len(batch) > 0:
+        if len(batch) > 0 and whole_blob_read:
             result.extend(func(batch, params))
-            
-        return result, whole_blob_read
+
+        return result, whole_blob_read, blobs_iterator
 
     def _record_failure(self, request_id, e: Exception):
         track = traceback.format_exc()
